"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.EufySecurityHomebridgePlatform = void 0;
const eufy_node_client_1 = require("eufy-node-client");
const settings_1 = require("./settings");
const doorbell_platform_accessory_1 = require("./doorbell-platform-accessory");
const eufy_types_1 = require("./eufy-types");
const fs_1 = __importDefault(require("fs"));
/**
 * HomebridgePlatform
 * This class is the main constructor for your plugin, this is where you should
 * parse the user config and discover/register accessories with Homebridge.
 */
class EufySecurityHomebridgePlatform {
    constructor(log, config, api) {
        this.log = log;
        this.config = config;
        this.api = api;
        this.Service = this.api.hap.Service;
        this.Characteristic = this.api.hap.Characteristic;
        // this is used to track restored cached accessories
        this.accessories = [];
        this.log.debug('Config', this.config);
        this.log.debug('Finished initializing platform:', this.config.name);
        this.httpService = new eufy_node_client_1.HttpService(config.username, config.password);
        // When this event is fired it means Homebridge has restored all cached accessories from disk.
        // Dynamic Platform plugins should only register new accessories after this event was fired,
        // in order to ensure they weren't added to homebridge already. This event can also be used
        // to start discovery of new accessories.
        this.api.on('didFinishLaunching', async () => {
            await this.setupPushClient();
            log.debug('Executed didFinishLaunching callback');
            // run the method to discover / register your devices as accessories
            this.discoverDevices();
        });
    }
    async setupPushClient() {
        const storagePath = this.api.user.storagePath();
        const credentialsPath = `${storagePath}/eufy-security-credentials.json`;
        let credentials = null;
        if (fs_1.default.existsSync(credentialsPath)) {
            this.log.info('credentials found. reusing them...');
            credentials = JSON.parse(fs_1.default.readFileSync(credentialsPath).toString());
        }
        else {
            // Register push credentials
            this.log.info('no credentials found. register new...');
            const pushService = new eufy_node_client_1.PushRegisterService();
            credentials = await pushService.createPushCredentials();
            fs_1.default.writeFileSync(credentialsPath, JSON.stringify(credentials));
            this.log.info('wait a short time (5sec)...');
            await new Promise(r => setTimeout(r, 5000));
        }
        // Start push client
        const pushClient = await eufy_node_client_1.PushClient.init({
            androidId: credentials.checkinResponse.androidId,
            securityToken: credentials.checkinResponse.securityToken,
        });
        const fcmToken = credentials.gcmResponse.token;
        await this.httpService.registerPushToken(fcmToken);
        this.log.debug('Registered at eufy with:', fcmToken);
        setInterval(async () => {
            await this.httpService.pushTokenCheck();
        }, 30 * 1000);
        pushClient.connect((msg) => {
            this.log.debug('Got push message:', msg);
            const matchingUuid = this.api.hap.uuid.generate(msg.payload.device_sn);
            const knownAccessory = this.accessories.find(accessory => accessory.UUID === matchingUuid);
            if (knownAccessory) {
                knownAccessory.getService(this.api.hap.Service.Doorbell)
                    .updateCharacteristic(this.api.hap.Characteristic.ProgrammableSwitchEvent, this.api.hap.Characteristic.ProgrammableSwitchEvent.SINGLE_PRESS);
            }
            this.log.info('Got push message:', msg);
        });
    }
    /**
     * This function is invoked when homebridge restores cached accessories from disk at startup.
     * It should be used to setup event handlers for characteristics and update respective values.
     */
    configureAccessory(accessory) {
        this.log.info('Loading accessory from cache:', accessory.displayName);
        // add the restored accessory to the accessories cache so we can track if it has already been registered
        this.accessories.push(accessory);
    }
    /**
     * This is an example method showing how to register discovered accessories.
     * Accessories must only be registered once, previously created accessories
     * must not be registered again to prevent "duplicate UUID" errors.
     */
    async discoverDevices() {
        var _a, _b;
        const hubs = await this.httpService.listHubs();
        for (const hub of hubs) {
            if ((_a = this.config.ignoreHubSns) === null || _a === void 0 ? void 0 : _a.includes(hub.station_sn)) {
                continue;
            }
            const { station_sn } = hub;
            const devices = await this.httpService.listDevices({ station_sn });
            for (const device of devices) {
                if ((_b = this.config.ignoreDeviceIds) === null || _b === void 0 ? void 0 : _b.includes(device.device_sn)) {
                    continue;
                }
                const uuid = this.api.hap.uuid.generate(device.device_sn);
                const existingAccessory = this.accessories.find(accessory => accessory.UUID === uuid);
                // doorbell
                if (device.device_type === eufy_types_1.DeviceType.BATTERY_DOORBELL) {
                    if (existingAccessory) {
                        // the accessory already exists
                        this.log.info('Restoring existing accessory from cache:', existingAccessory.displayName);
                        new doorbell_platform_accessory_1.DoorbellPlatformAccessory(this, existingAccessory, device);
                    }
                    else {
                        // the accessory does not yet exist, so we need to create it
                        this.log.info('Adding new accessory:', device.device_name);
                        // create a new accessory
                        const accessory = new this.api.platformAccessory(device.device_name, uuid);
                        // store a copy of the device object in the `accessory.context`
                        // the `context` property can be used to store any data about the accessory you may need
                        accessory.context.device = device;
                        // create the accessory handler for the newly create accessory
                        // this is imported from `platformAccessory.ts`
                        new doorbell_platform_accessory_1.DoorbellPlatformAccessory(this, accessory, device);
                        // link the accessory to your platform
                        this.api.registerPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [accessory]);
                    }
                }
            }
        }
        // @todo
        // this.api.unregisterPlatformAccessories(PLUGIN_NAME, PLATFORM_NAME, [existingAccessory]);
        // this.log.info('Removing existing accessory from cache:', existingAccessory.displayName);
    }
}
exports.EufySecurityHomebridgePlatform = EufySecurityHomebridgePlatform;
//# sourceMappingURL=platform.js.map